package example;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assume.assumeFalse;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URL;
import java.util.Collections;

import org.eclipse.core.runtime.preferences.IEclipsePreferences;
import org.eclipse.core.runtime.preferences.InstanceScope;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.uml2.uml.Class;
import org.eclipse.uml2.uml.Model;
import org.eclipse.uml2.uml.Package;
import org.eclipse.uml2.uml.Property;
import org.eclipse.uml2.uml.UMLPackage;
import org.eclipse.uml2.uml.util.UMLUtil;
import org.junit.FixMethodOrder;
import org.junit.Test;
import org.junit.runners.MethodSorters;
import org.osgi.framework.Bundle;
import org.osgi.framework.BundleReference;

import com.yakindu.bridges.ea.core.EAResource;
import com.yakindu.bridges.ea.uml.utils.EAResourceUMLUtils;

@FixMethodOrder(MethodSorters.NAME_ASCENDING)
public class LoadAndSaveModelTest {

	@Test
	public void should_LoadExample() throws IOException {
		// given
		File tempFile = extractToTemp("model/ExampleWithMDGTechnology.eap");
		URI uri = URI.createFileURI(tempFile.getAbsolutePath());
		// when
		Resource resource = openResource(uri, false);
		try {
			// then
			Model model = (Model) resource.getContents().get(0);
			Package view = (Package) model.getPackagedElement("View");
			Class clazz = (Class) view.getPackagedElement("TheClass");
			Property attribute = clazz.getOwnedAttributes().get(0);
			assertEquals("theAttribute", attribute.getName());
			assertEquals("TheMDGDatatype", attribute.getType().getName());
			assertNotNull(clazz.getAppliedStereotype("TheMDGTechnology::TheStereotype"));
			assertEquals("theTaggedValueValue",
					UMLUtil.getTaggedValue(clazz, "TheMDGTechnology::TheStereotype", "theTaggedValue"));
		} finally {
			close(resource);
		}
	}

	@Test
	public void should_SaveExample() throws IOException {
		// given
		File tempFile = extractToTemp("model/ExampleWithMDGTechnology.eap");
		URI uri = URI.createFileURI(tempFile.getAbsolutePath());
		Resource resource = openResource(uri, true);
		assumeFalse("Resource is still read-only, maybe because it's a trial version.",
				((EAResource) resource).isReadonly());
		try {
			Model model = (Model) resource.getContents().get(0);
			Package view = (Package) model.getPackagedElement("View");
			// when
			view.createOwnedType("NewClass", UMLPackage.eINSTANCE.getClass_());
			resource.save(Collections.emptyMap());
		} finally {
			close(resource);
		}
		// then
		Resource reloadedResource = openResource(uri, true);
		try {
			Model model = (Model) reloadedResource.getContents().get(0);
			Package view = (Package) model.getPackagedElement("View");
			Class clazz = (Class) view.getPackagedElement("NewClass");
			assertNotNull(clazz);
		} finally {
			close(reloadedResource);
		}
	}

	@Test
	public void should_SaveExample_ReadonlyDisabledInPreferences() throws IOException {
		// given
		IEclipsePreferences node = InstanceScope.INSTANCE.getNode(EAResource.EA_PREFERENCES_NODE);
		boolean readonlyByDefault = node.getBoolean(EAResource.OPTION_READONLY, EAResource.OPTION_READONLY_DEFAULT);
		try {
			node.putBoolean(EAResource.OPTION_READONLY, false);
			File tempFile = extractToTemp("model/ExampleWithMDGTechnology.eap");
			URI uri = URI.createFileURI(tempFile.getAbsolutePath());
			Resource resource = openResource(uri, false);
			assumeFalse("Resource is still read-only, maybe because it's a trial version.",
					((EAResource) resource).isReadonly());
			try {
				Model model = (Model) resource.getContents().get(0);
				Package view = (Package) model.getPackagedElement("View");
				// when
				view.createOwnedType("NewClass", UMLPackage.eINSTANCE.getClass_());
				resource.save(Collections.emptyMap());
			} finally {
				close(resource);
			}
			// then
			Resource reloadedResource = openResource(uri, false);
			try {
				Model model = (Model) reloadedResource.getContents().get(0);
				Package view = (Package) model.getPackagedElement("View");
				Class clazz = (Class) view.getPackagedElement("NewClass");
				assertNotNull(clazz);
			} finally {
				close(reloadedResource);
			}
		} finally {
			node.putBoolean(EAResource.OPTION_READONLY, readonlyByDefault);
		}
	}

	protected Resource openResource(URI uri, boolean disableReadonly) {
		ResourceSet rst = new ResourceSetImpl();
		if (disableReadonly) {
			rst.getLoadOptions().put(EAResource.OPTION_READONLY, false);
		}
		Resource resource = rst.getResource(uri, true);
		assertEquals(0, resource.getErrors().size());
		assertEquals(0, resource.getWarnings().size());
		return resource;
	}

	@SuppressWarnings("restriction")
	protected void close(final Resource resource) {
		if (resource != null) {
			resource.unload();
			if (EAResourceUMLUtils.isLocalEaResource(resource)) {
				// Force JET driver to close file handles. It might help to improve test
				// stability. Not recommended in productive code!
				com.yakindu.bridges.ea.core.internal.jdbc.HXTTAccess.releaseLocks();
			}
		}
	}

	protected File extractToTemp(String pathInBundle) throws IOException {
		Bundle bundle = getBundle();
		assertNotNull(
				"This project is not available as a plug-in, ensure that tests are exeucted as JUnit Plug-in Test!",
				bundle);
		File tempFile = File.createTempFile("test", "." + EAResource.FILE_EXTENSION);
		tempFile.deleteOnExit();
		URL url = bundle.getEntry(pathInBundle);
		try (InputStream source = url.openStream()) {
			try (OutputStream target = new FileOutputStream(tempFile)) {
				final byte[] buffer = new byte[4096];
				int read;
				while ((read = source.read(buffer)) > 0) {
					target.write(buffer, 0, read);
				}
			}
		}
		return tempFile;
	}

	private Bundle getBundle() {
		ClassLoader classLoader = getClass().getClassLoader();
		return classLoader instanceof BundleReference ? ((BundleReference) classLoader).getBundle() : null;
	}
}
